class PlpygisError(Exception):
    """
    Basic exception for ``plpygis``.
    """

    def __init__(self, msg):
        super().__init__(msg)


class CoordinateError(PlpygisError):
    """
    Exception for problems in the coordinates of geometries.
    """

    def __init__(self, geom, msg=None):
        if msg is None:
            msg = f"Geometry has invalid coordinates: {geom}"
        super().__init__(msg)


class CollectionError(PlpygisError):
    """
    Exception for problems with geometries in collection types.
    """

    def __init__(self, msg=None):
        if msg is None:
            msg = "Error in the geometries in a collection."
        super().__init__(msg)


class DependencyError(PlpygisError, ImportError):
    """
    Exception for a missing dependency.
    """

    def __init__(self, dep, msg=None):
        if msg is None:
            msg = f"Dependency '{dep}' is not available."
        super().__init__(msg)


class WkbError(PlpygisError):
    """
    Exception for problems in parsing WKBs.
    """

    def __init__(self, msg=None):
        if msg is None:
            msg = "Unreadable WKB."
        super().__init__(msg)


class WktError(Exception):
    """
    Exception for problems in parsing WKTs.
    """

    def __init__(self, reader, msg=None, expected=None):
        pos = reader.pos
        if not msg:
            if expected is None:
                msg = f"Unreadable WKT at position {pos+1}."
            else:
                msg = f"Expected {expected} at position {pos+1}."
        super().__init__(msg)


class DimensionalityError(PlpygisError):
    """
    Exception for problems in dimensionality of geometries.
    """

    def __init__(self, msg=None):
        if msg is None:
            msg = "Geometry has invalid dimensionality."
        super().__init__(msg)


class SridError(PlpygisError):
    """
    Exception for problems in dimensionality of geometries.
    """

    def __init__(self, msg=None):
        if msg is None:
            msg = "Geometry has invalid SRID."
        super().__init__(msg)


class GeojsonError(PlpygisError):
    """
    Exception for problems in GeoJSONs.
    """

    def __init__(self, msg=None):
        if msg is None:
            msg = "Invalid GeoJSON."
        super().__init__(msg)



"""
Test Geometry
"""

import pytest
from shapely import geometry
from plpygis import Geometry, Point, LineString, Polygon
from plpygis import MultiPoint, MultiLineString, MultiPolygon, GeometryCollection
from plpygis.exceptions import WkbError, SridError, DimensionalityError, CoordinateError, GeojsonError, CollectionError, WktError
from copy import copy, deepcopy

geojson_pt = {"type":"Point","coordinates":[0.0,0.0]}
geojson_ln = {"type":"LineString","coordinates":[[107,60],[102,59]]}
geojson_pg = {"type":"Polygon","coordinates":[[[100,0],[101.0,0.0],[101.0,1.0],[100.0,1.0],[100.0,0.0]]]}
geojson_mpt = {"type":"MultiPoint","coordinates":[[0,0],[1,1]]}
geojson_mln = {"type":"MultiLineString","coordinates":[[[0,0],[1,1]],[[2,2],[3,3]]]}
geojson_mpg = {"type":"MultiPolygon","coordinates":[[[[1,0],[111,0.0],[101.0,1.0],[100.0,1.0],[1,0]]],[[[100,0],[101.0,0.0],[101.0,1.0],[100.0,1.0],[100.0,0.0]]]]}
geojson_gc = {"type":"GeometryCollection","geometries":[{"type":"Point","coordinates":[10,0]},{"type":"LineString","coordinates":[[11,0],[12,1]]}]}
geojson_err = {"type":"Hello","coordinates":[0.0,0.0]}
geojson_pg_ring = {"type":"Polygon","coordinates":[[[100,0], [101,0],[201,1],[100,1],[100,0]],[[100.2,0.2],[100.8,0.2],[100.8,0.8],[100.2,0.8],[100.2,0.2]]]}
wkb_ln = "0102000000050000000000000040BE40409D640199EB373F400000000080AC3E40BF244710FD1939400000000000503940D2A6484BEB41374000000000801D3740248729C89C832A400000000000833340940338EFAFBB2C40"
wkb_pg = "010300000002000000060000000000000000003440000000000080414000000000000024400000000000003E40000000000000244000000000000024400000000000003E4000000000000014400000000000804640000000000000344000000000000034400000000000804140040000000000000000003E40000000000000344000000000000034400000000000002E40000000000000344000000000000039400000000000003E400000000000003440"
wkb_mpt = "010400008002000000010100008000000000000059400000000000006940000000000000000001010000800000000000000000000000000000F03F0000000000000000"
wkb_mln = "010500004002000000010200004002000000000000000000000000000000000000000000000000004940000000000000F03F000000000000F03F0000000000003940010200004002000000000000000000F0BF000000000000F0BF000000000000F03F2DB29DEFA7C60140ED0DBE3099AA0A400000000000388F40"
wkb_mpg = "01060000000200000001030000000100000004000000000000000000444000000000000044400000000000003440000000000080464000000000008046400000000000003E4000000000000044400000000000004440010300000002000000060000000000000000003440000000000080414000000000000024400000000000003E40000000000000244000000000000024400000000000003E4000000000000014400000000000804640000000000000344000000000000034400000000000804140040000000000000000003E40000000000000344000000000000034400000000000002E40000000000000344000000000000039400000000000003E400000000000003440"
wkb_gc = "0107000000020000000101000000000000000000000000000000000000000102000000020000000000000000000000000000000000F03F000000000000F03F000000000000F03F"
wkb_mpt_srid = "0104000020e8030000020000000101000000000000000000000000000000000000000101000000000000000000f03f000000000000f03f"

def test_missing_ewkb():
    """
    Error on missing EWKB
    """
    with pytest.raises(WkbError):
        Geometry(None)

def test_malformed_ewkb_len():
    """
    malformed EWKB (insufficient bytes)
    """
    with pytest.raises(Exception):
        Geometry("0101")

def test_wkb_type():
    """
    malformed EWKB (insufficient bytes)
    """
    with pytest.raises(WkbError):
        Geometry(0)

def test_malformed_ewkb_firstbyte():
    """
    malformed EWKB (bad first byte)
    """
    with pytest.raises(WkbError):
        Geometry("5101")

def test_unsupported_ewkb_type():
    """
    unsupported EWKB type
    """
    with pytest.raises(WkbError):
        Geometry("010800000000000000000000000000000000000000")

def test_read_wkb_point():
    """
    read WKB Point
    """
    wkb = "010100000000000000000000000000000000000000"
    geom = Geometry(wkb)
    assert geom.type == "Point"
    assert geom.srid is None
    assert not geom.dimz
    assert not geom.dimm
    postgis_type = "geometry(Point)"
    assert geom.postgis_type == postgis_type
    assert geom.__repr__() == "<Point: 'geometry(Point)'>"
    geom.srid = geom.srid  # clear cached WKB
    assert geom.__str__().lower() == wkb.lower()

def test_read_wkb_point_big_endian():
    """
    read WKB Point
    """
    geom = Geometry("000000000140000000000000004010000000000000")
    assert isinstance(geom, Point)
    assert geom.x == 2
    assert geom.y == 4
    assert geom.z is None

def test_read_ewkb_point_srid():
    """
    read EWKB Point,4326
    """
    wkb = "0101000020E610000000000000000000000000000000000000"
    geom = Geometry(wkb)
    assert geom.type == "Point"
    assert geom.srid == 4326
    assert not geom.dimz
    assert not geom.dimm
    postgis_type = "geometry(Point,4326)"
    assert geom.postgis_type == postgis_type
    assert geom.__repr__() == "<Point: 'geometry(Point,4326)'>"
    geom.srid = geom.srid # clear cached WKB
    assert geom.ewkb == wkb.lower()
    assert geom.wkb != geom.ewkb
    assert geom.__str__().lower() == wkb.lower()

def test_read_ewkb_pointz():
    """
    read EWKB PointZ,4326
    """
    wkb = "01010000A0E6100000000000000000000000000000000000000000000000000000"
    geom = Geometry(wkb)
    assert geom.type == "Point"
    assert geom.srid == 4326
    assert geom.dimz
    assert not geom.dimm
    postgis_type = "geometry(PointZ,4326)"
    assert geom.postgis_type == postgis_type
    assert geom.__repr__() == "<Point: 'geometry(PointZ,4326)'>"
    geom.srid = geom.srid # clear cached WKB
    assert geom.ewkb == wkb.lower()
    assert geom.wkb != geom.ewkb
    assert geom.__str__().lower() == wkb.lower()

def test_read_ewkb_pointm():
    """
    read EWKB PointM,4326
    """
    wkb = "0101000060E6100000000000000000000000000000000000000000000000000000"
    geom = Geometry(wkb)
    assert geom.type == "Point"
    assert geom.srid == 4326
    assert not geom.dimz
    assert geom.dimm
    postgis_type = "geometry(PointM,4326)"
    assert geom.postgis_type == postgis_type
    assert geom.__repr__() == "<Point: 'geometry(PointM,4326)'>"
    geom.srid = geom.srid # clear cached WKB
    assert geom.ewkb == wkb.lower()
    assert geom.wkb != geom.ewkb
    assert geom.__str__().lower() == wkb.lower()

def test_read_ewkb_pointzm():
    """
    read EWKB PointZM,4326
    """
    wkb = "01010000E0E61000000000000000000000000000000000000000000000000000000000000000000000"
    geom = Geometry(wkb)
    assert geom.type == "Point"
    assert geom.srid == 4326
    assert geom.dimz
    assert geom.dimm
    geom.srid = geom.srid # clear cached WKB
    postgis_type = "geometry(PointZM,4326)"
    assert geom.postgis_type == postgis_type
    assert geom.__repr__() == "<Point: 'geometry(PointZM,4326)'>"
    geom.srid = geom.srid # clear cached WKB
    assert geom.ewkb == wkb.lower()
    assert geom.wkb != geom.ewkb
    assert geom.__str__().lower() == wkb.lower()

def test_read_wkb_data_error():
    """
    read WKB with good header but malformed data
    """
    wkb = "0000000001000000000000"
    geom = Geometry(wkb)
    assert geom.type == "Point"
    with pytest.raises(WkbError):
        geom.x

def test_read_wkb_linestring():
    """
    read WKB LineString
    """
    wkb = wkb_ln
    geom = Geometry(wkb)
    assert geom.type == "LineString"
    assert geom.srid is None
    assert not geom.dimz
    assert not geom.dimm
    postgis_type = "geometry(LineString)"
    geom.vertices
    assert geom.postgis_type == postgis_type
    assert geom.__repr__() == "<LineString: 'geometry(LineString)'>"
    geom.srid = geom.srid # clear cached WKB
    assert geom.__str__().lower() == wkb.lower()

def test_read_wkb_polygon():
    """
    read WKB Polygon
    """
    wkb = wkb_pg
    geom = Geometry(wkb)
    assert geom.type == "Polygon"
    assert geom.srid is None
    assert not geom.dimz
    assert not geom.dimm
    postgis_type = "geometry(Polygon)"
    assert geom.exterior.type == "LineString"
    assert geom.postgis_type == postgis_type
    assert geom.__repr__() == "<Polygon: 'geometry(Polygon)'>"
    geom.srid = geom.srid # clear cached WKB
    assert geom.__str__().lower() == wkb.lower()

def test_read_wkb_multipoint():
    """
    read WKB MultiPoint
    """
    wkb = wkb_mpt
    geom = Geometry(wkb)
    assert geom.type == "MultiPoint"
    assert geom.srid is None
    assert geom.dimz
    assert not geom.dimm
    postgis_type = "geometry(MultiPointZ)"
    assert geom.postgis_type == postgis_type
    assert geom.__repr__() == "<MultiPoint: 'geometry(MultiPointZ)'>"
    geom.srid = geom.srid # clear cached WKB
    assert geom.__str__().lower() == wkb.lower()
    for g in geom.geometries:
        assert g.type == "Point"
    for g in geom.points:
        assert g.type == "Point"
    assert wkb == geom.wkb

def test_read_wkb_multilinestring():
    """
    read WKB MultiLineString
    """
    wkb = wkb_mln
    geom = Geometry(wkb)
    assert geom.type == "MultiLineString"
    assert geom.srid is None
    assert not geom.dimz
    assert geom.dimm
    postgis_type = "geometry(MultiLineStringM)"
    assert geom.postgis_type == postgis_type
    assert geom.__repr__() == "<MultiLineString: 'geometry(MultiLineStringM)'>"
    geom.srid = geom.srid # clear cached WKB
    assert geom.__str__().lower() == wkb.lower()
    for g in geom.geometries:
        assert g.type == "LineString"
    for g in geom.linestrings:
        assert g.type == "LineString"
    assert wkb == geom.wkb

def test_read_wkb_multipolygon():
    """
    read WKB MultiPolygon
    """
    wkb = wkb_mpg
    geom = Geometry(wkb)
    assert geom.type == "MultiPolygon"
    assert geom.srid is None
    assert not geom.dimz
    assert not geom.dimm
    postgis_type = "geometry(MultiPolygon)"
    assert geom.postgis_type == postgis_type
    assert geom.__repr__() == "<MultiPolygon: 'geometry(MultiPolygon)'>"
    geom.srid = geom.srid # clear cached WKB
    assert geom.__str__().lower() == wkb.lower()
    for g in geom.geometries:
        assert g.type == "Polygon"
    for g in geom.polygons:
        assert g.type == "Polygon"
    assert wkb == geom.wkb

def test_read_wkb_geometrycollection():
    """
    read WKB GeometryCollection
    """
    wkb = wkb_gc
    geom = Geometry(wkb)
    assert geom.type == "GeometryCollection"
    assert geom.srid is None
    assert not geom.dimz
    assert not geom.dimm
    postgis_type = "geometry(GeometryCollection)"
    assert geom.postgis_type == postgis_type
    assert geom.__repr__() == "<GeometryCollection: 'geometry(GeometryCollection)'>"
    geom.srid = geom.srid # clear cached WKB
    assert geom.__str__().lower() == wkb.lower()
    assert geom.geometries[0].type == "Point"
    assert geom.geometries[1].type == "LineString"
    assert wkb == geom.wkb

def test_multigeometry_raise_error():
    """
    raise error when adding wrong type to a multigeometry
    """
    pt = Point((0,1))
    ls = LineString([(0,1), (2,3)])

    MultiPoint([pt, copy(pt)])

    with pytest.raises(CollectionError):
        MultiPoint([pt, ls])

    with pytest.raises(CollectionError):
        MultiLineString([pt, ls])

    with pytest.raises(CollectionError):
        MultiPolygon([pt, ls])

    with pytest.raises(CollectionError):
        GeometryCollection([pt, ls, True])

def test_multigeometry_nochangedimensionality():
    mp = MultiPoint([Point((0, 1, 2)), Point((5, 6, 7))])
    mp.dimz = True
    assert mp.dimz == True
    mp.dimm = False
    assert mp.dimm == False

def test_point_nullifyzm():
    p = Point((0, 1, 2, 3))
    assert p.dimz == True
    p.z = None
    assert p.dimz == False
    p.dimz = False
    assert p.dimz == False

    assert p.dimm == True
    p.m = None
    assert p.dimm == False
    p.dimm = False
    assert p.dimm == False

def test_multigeometry_changedimensionality():
    """
    change dimensionality of a MultiGeometry
    """
    wkb = wkb_gc
    geom = Geometry(wkb)
    assert not geom.dimz
    assert not geom.dimm
    geom.dimz = True
    geom.dimm = True
    assert geom.dimz
    assert geom.dimm
    geom.srid = geom.srid # clear cached WKB
    assert geom.__str__().lower() != wkb.lower()

def test_malformed_coordinates():
    """
    malformed coordinates (wrong type)
    """
    coordinates = (1, "test")
    with pytest.raises(CoordinateError):
        Point(coordinates)

    coordinates = [(1, 2), [(1, 2), (3, 4)]]
    with pytest.raises(CoordinateError):
        LineString(coordinates)

def test_multigeometry_srid():
    """
    create geometry with SRID
    """
    p1 = Point((0, 0), srid=1000)
    p2 = Point((1, 1), srid=1000)
    mp = MultiPoint([p1, p2], srid=1000)
    assert mp.ewkb == wkb_mpt_srid

def test_multigeometry_srid_exception():
    """
    mixed SRIDs on multigeometry creation
    """
    p1 = Point((0, 0), srid=1000)
    p2 = Point((1, 1), srid=1000)
    with pytest.raises(SridError):
        MultiPoint([p1, p2])

    p1 = Point((0, 0), srid=1000)
    p2 = Point((1, 1), srid=1000)
    with pytest.raises(SridError):
        MultiPoint([p1, p2], 4326)

def test_translate_geojson_pt():
    """
    load and dump GeoJSON point
    """
    geom = Geometry.from_geojson(geojson_pt)
    assert geom.srid == 4326
    assert Point == type(geom)
    geojson = geom.geojson
    assert geojson == geojson_pt

def test_translate_geojson_ln():
    """
    load and dump GeoJSON line
    """
    geom = Geometry.from_geojson(geojson_ln)
    assert geom.srid == 4326
    assert LineString == type(geom)
    geojson = geom.geojson
    assert geojson == geojson_ln

def test_translate_geojson_pg():
    """
    load and dump GeoJSON polygon
    """
    geom = Geometry.from_geojson(geojson_pg)
    assert geom.srid == 4326
    assert Polygon == type(geom)
    geojson = geom.geojson
    assert geojson == geojson_pg

def test_translate_geojson_mpt():
    """
    load and dump GeoJSON multipoint
    """
    geom = Geometry.from_geojson(geojson_mpt)
    assert geom.srid == 4326
    assert MultiPoint == type(geom)
    geojson = geom.geojson
    assert geojson == geojson_mpt

def test_translate_geojson_mln():
    """
    load and dump GeoJSON multiline
    """
    geom = Geometry.from_geojson(geojson_mln)
    assert geom.srid == 4326
    assert MultiLineString == type(geom)
    geojson = geom.geojson
    assert geojson == geojson_mln

def test_translate_geojson_mpg():
    """
    load and dump GeoJSON multipolygon
    """
    geom = Geometry.from_geojson(geojson_mpg)
    assert geom.srid == 4326
    assert MultiPolygon == type(geom)
    geojson = geom.geojson
    assert geojson == geojson_mpg

def test_translate_geojson_gc():
    """
    load and dump GeoJSON GeometryCollection
    """
    geom = Geometry.from_geojson(geojson_gc)
    assert geom.srid == 4326
    assert GeometryCollection == type(geom)
    geojson = geom.geojson
    assert geojson == geojson_gc

def test_translate_geojson_error():
    """
    catch invalid GeoJSON
    """
    with pytest.raises(GeojsonError):
        Geometry.from_geojson(geojson_err)

def test_geo_interface():
    """
    access using __geo_interface__
    """
    geom = Geometry.from_geojson(geojson_pt)
    assert geojson_pt == geom.__geo_interface__

def test_shape():
    """
    access using __geo_interface__ and shape
    """
    point = Point.from_geojson(geojson_pt)
    geom = Geometry.shape(point)
    assert point.__geo_interface__ == geom.__geo_interface__

def test_shapely_dump():
    """
    convert to Shapely
    """
    point = Point((123,123))
    sgeom = point.shapely
    assert sgeom.wkb == point.wkb

def test_shapely_load():
    """
    convert from Shapely
    """
    sgeom = geometry.Point(99,-99)
    point = Geometry.from_shapely(sgeom)
    assert sgeom.wkb_hex == point.wkb

def test_strip_srid():
    """
    strip SRID
    """
    geom1 = Geometry("010100000000000000000000000000000000000000")
    geom2 = Geometry("0101000020E610000000000000000000000000000000000000")
    assert geom1.srid is None
    assert geom2.srid == 4326
    geom2.srid = None
    assert geom2.srid is None
    assert geom1.wkb == geom2.wkb
    geom2.srid = None
    assert geom2.srid is None
    assert geom1.wkb == geom2.wkb

def test_bounds_point():
    """
    check bounds of Point
    """
    geom = Geometry.from_geojson(geojson_pt)
    bounds = geom.bounds
    assert bounds[0] == 0.0
    assert bounds[1] == 0.0
    assert bounds[2] == 0.0
    assert bounds[3] == 0.0

def test_bounds_linestring():
    """
    check bounds of LineString
    """
    geom = Geometry.from_geojson(geojson_ln)
    bounds = geom.bounds
    assert bounds[0] == 102
    assert bounds[1] == 59
    assert bounds[2] == 107
    assert bounds[3] == 60

def test_bounds_polygon():
    """
    check bounds of Polygon
    """
    geom = Geometry.from_geojson(geojson_pg)
    bounds = geom.bounds
    assert bounds[0] == 100
    assert bounds[1] == 0
    assert bounds[2] == 101
    assert bounds[3] == 1

def test_bounds_multipoint():
    """
    check bounds of MultiPoint
    """
    geom = Geometry.from_geojson(geojson_mpt)
    bounds = geom.bounds
    assert bounds[0] == 0
    assert bounds[1] == 0
    assert bounds[2] == 1
    assert bounds[3] == 1

def test_bounds_multilinestring():
    """
    check bounds of MultiLineString
    """
    geom = Geometry.from_geojson(geojson_mln)
    bounds = geom.bounds
    assert bounds[0] == 0
    assert bounds[1] == 0
    assert bounds[2] == 3
    assert bounds[3] == 3

def test_bounds_multipolygon():
    """
    check bounds of MultiPolygon
    """
    geom = Geometry.from_geojson(geojson_mpg)
    bounds = geom.bounds
    assert bounds[0] == 1
    assert bounds[1] == 0
    assert bounds[2] == 111
    assert bounds[3] == 1

def test_bounds_geomcollection():
    """
    check bounds of GeometryCollection
    """
    geom = Geometry.from_geojson(geojson_gc)
    bounds = geom.bounds
    assert bounds[0] == 10
    assert bounds[1] == 0
    assert bounds[2] == 12
    assert bounds[3] == 1

def test_mixed_dimensionality():
    """
    detect mixed dimensionality in MultiGeometries
    """
    p1 = Point((0, 1, 2))
    p2 = Point((0, 1))
    with pytest.raises(DimensionalityError):
        MultiPoint([p1, p2])

def test_invalid_point_dimension():
    """
    detect extra dimensions in Point creation
    """
    with pytest.raises(DimensionalityError):
        Point((0, 1, 2, 3, 4))

def test_dimension_reading():
    """
    check dimensions are set correctly
    """
    p = Point((0, 1))
    assert not p.dimz
    assert not p.dimm
    p = Point((0, 1, 2))
    assert p.dimz
    assert not p.dimm
    p = Point((0, 1, 2, 3))
    assert p.dimz
    assert p.dimm
    p = Point((0, 1, 2, 3), dimz=False, dimm=False)
    assert p.dimz
    assert p.dimm
    p = Point((0, 1), dimz=True, dimm=True)
    assert p.dimz
    assert p.dimm
    p = Point((0, 1), dimz=True)
    assert p.dimz
    assert not p.dimm
    p = Point((0, 1), dimm=True)
    assert not p.dimz
    assert p.dimm
    p = Point((0, 1, 2), dimz=True)
    assert p.dimz
    assert not p.dimm
    p = Point((0, 1, 2), dimm=True)
    assert not p.dimz
    assert p.dimm
    p = Point((0, 1, 2), dimz=True, dimm=True)
    assert p.dimz
    assert p.dimm

def test_modify_point():
    """
    modify Point
    """
    wkb = "010100000000000000000000000000000000000000"
    p = Geometry(wkb)
    oldx = p.x
    oldy = p.y
    oldsrid = p.srid
    assert not p.dimz
    assert not p.dimm
    newx = -99
    newy = -101
    newz = 88
    newm = 8
    newsrid = 900913
    assert p.x != newx
    assert p.y != newy
    assert p.z != newz
    assert p.m != newm
    assert p.srid != newsrid
    p.x = newx
    p.y = newy
    p.z = newz
    p.m = newm
    p.srid = newsrid
    assert p.x == newx
    assert p.y == newy
    assert p.z == newz
    assert p.m == newm
    assert p.srid == newsrid
    assert p.__str__().lower() != wkb.lower()
    p.x = oldx
    p.y = oldy
    p.srid = oldsrid
    p.dimz = None
    p.dimm = None
    assert p.__str__().lower() == wkb.lower()

def test_binary_wkb_roundtrip():
    wkb = b'\x01\x01\x00\x00\x20\xe6\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    assert Geometry(wkb).wkb == wkb

def test_byte_array_wkb_string():
    wkb = bytearray(b'010100000000000000000000000000000000000000')
    assert str(Geometry(wkb)) == wkb.decode('ascii')

def test_point_coordinates():
    """
    get coordinates of a Point
    """
    coordinates = (1, 2, 3, 4)
    p = Point(coordinates)
    assert p.coordinates == coordinates
    assert p._coordinates(dimm=True, dimz=True) == coordinates
    assert p._coordinates(dimm=False) == (1, 2, 3)
    assert p._coordinates(dimz=False) == (1, 2, 4)
    assert p._coordinates(dimm=False, dimz=False) == (1, 2)

    assert p._coordinates(dimm=True, dimz=True, tpl=False) == list(coordinates)
    assert p._coordinates(dimm=False, tpl=False) == [1, 2, 3]
    assert p._coordinates(dimz=False, tpl=False) == [1, 2, 4]
    assert p._coordinates(dimm=False, dimz=False, tpl=False) == [1, 2]

def test_linestring_coordinates():
    """
    get coordinates of a LineString
    """
    coordinates = [(1, 2, 3, 4), (6, 7, 8, 9)]
    ls = LineString(coordinates)
    assert ls.coordinates == coordinates
    assert ls._coordinates(dimm=True, dimz=True) == coordinates
    assert ls._coordinates(dimm=False) == [(1, 2, 3), (6, 7, 8)]
    assert ls._coordinates(dimz=False) == [(1, 2, 4), (6, 7, 9)]
    assert ls._coordinates(dimz=False, dimm=False) == [(1, 2), (6, 7)]

    assert ls._coordinates(dimm=True, dimz=True, tpl=False) == [[1, 2, 3, 4], [6, 7, 8, 9]]
    assert ls._coordinates(dimm=False, tpl=False) == [[1, 2, 3], [6, 7, 8]]
    assert ls._coordinates(dimz=False, tpl=False) == [[1, 2, 4], [6, 7, 9]]
    assert ls._coordinates(dimz=False, dimm=False, tpl=False) == [[1, 2], [6, 7]]

def test_polygon_coordinates():
    """
    get coordinates of a Polygon
    """
    coordinates = [[(1, 2, 3, 4), (6, 7, 8, 9), (10, 11, 12, 13), (1, 2, 3, 4)]]
    p = Polygon(coordinates)
    assert p.coordinates == coordinates
    assert p._coordinates(dimm=True, dimz=True) == coordinates

    coordinates = [list(map(list, sub)) for sub in coordinates]
    assert p._coordinates(dimm=True, dimz=True, tpl=False) == coordinates

def test_multipoint_coordinates():
    """
    get coordinates of a MultiPoint
    """
    coordinates =  [ (170.0, 45.0), (180.0, 45.0), (-100.0, 45.0), (-170.0, 45.0) ]
    mp = MultiPoint([Point(p) for p in coordinates])
    assert mp.coordinates == coordinates

def test_multilinestring_coordinates():
    """
    get coordinates of a MultiLineString
    """
    coordinates = [ [ (170.0, 45.0), (180.0, 45.0) ], [ (-180.0, 45.0), (-170.0, 45.0) ] ]
    ml = MultiLineString([LineString(c) for c in coordinates])
    assert ml.coordinates == coordinates

def test_multipolygon_coordinates():
    """
    get coordinates of a MultiPolygon
    """
    coordinates = [ [ [ (180.0, 40.0), (180.0, 50.0), (170.0, 50.0), (170.0, 40.0), (180.0, 40.0) ] ], [ [ (-170.0, 40.0), (-170.0, 50.0), (-180.0, 50.0), (-180.0, 40.0), (-170.0, 40.0) ] ] ]
    mp = MultiPolygon([Polygon(c) for c in coordinates])
    assert mp.coordinates == coordinates

def test_point_copy():
    """
    copy a Point object
    """
    p1 = Point((0, 1, 2), srid=4326, dimm=True)
    p2 = copy(p1)
    p3 = copy(p1)

    assert p1.coordinates == p2.coordinates
    assert p1.srid == p2.srid
    assert p1.wkb == p2.wkb
    assert p1 != p2
    assert p1 != p3

    p1.x = 10
    assert p1.x != p2.x
    assert p2.x == p3.x

def test_multipoint_create():
    """
    create a MultiPoint object
    """
    p1 = Point((0, 1, 2))
    p2 = Point((3, 4, 5))
    p3 = Point((6, 7, 8))
    mp = MultiPoint([p1, p2, p3], srid=4326)

    p1.x = 100
    assert mp.geometries[0].x == p1.x
    assert p1.x == 100

def test_multipoint_copy():
    """
    copy a MultiPoint object
    """
    p1 = Point((0, 1, 2))
    p2 = Point((3, 4, 5))
    p3 = Point((6, 7, 8))
    mp1 = MultiPoint([p1, p2, p3], srid=4326)
    mp2 = copy(mp1)

    assert mp1.coordinates == mp2.coordinates

def test_geometrycollection_create():
    """
    create a GeometryCollection object
    """
    pt = Point((0, 1, 2))
    ls = LineString([(3, 4, 5), (9, 10, 11)])
    pl = Polygon([[(1, 2, 3), (6, 7, 8), (10, 11, 12), (1, 2, 3)]])
    gc1 = GeometryCollection([pt, ls, pl])
    gc2 = copy(gc1)
    for i, _ in enumerate(gc1):
        assert gc1[i] == gc2[i]
    pt.x = 100
    assert gc1.coordinates == gc2.coordinates

    gc1.geometries[0].x = 200
    assert gc1.coordinates == gc2.coordinates

def test_geometrycollection_copy():
    """
    modify a GeometryCollection object
    """
    pt = Point((0, 1, 2))
    ls = LineString([(3, 4, 5), (9, 10, 11)])
    pl = Polygon([[(1, 2, 3), (6, 7, 8), (10, 11, 12), (1, 2, 3)]])
    gc1 = GeometryCollection([pt, ls, pl])
    gc1.wkb
    gc2 = copy(gc1)
    assert gc1.wkb == gc2.wkb

    gc2.geometries[0].x = 123
    assert gc1.coordinates == gc2.coordinates
    assert gc2.geometries[0].x == 123
    assert gc1.wkb == gc2.wkb

    pt = Point((-1, -5, -1))
    gc2[1] = pt
    assert gc1.coordinates != gc2.coordinates
    assert gc2.geometries[1].x == -1
    assert gc2.geometries[1].y == -5
    assert gc1.wkb != gc2.wkb

def test_geometrycollection_deepcopy():
    """
    modify a GeometryCollection object
    """
    pt = Point((0, 1, 2))
    ls = LineString([(3, 4, 5), (9, 10, 11)])
    pl = Polygon([[(1, 2, 3), (6, 7, 8), (10, 11, 12), (1, 2, 3)]])
    gc1 = GeometryCollection([pt, ls, pl])
    gc1.wkb
    gc2 = deepcopy(gc1)
    assert gc1.wkb == gc2.wkb
    geoms = zip(gc1.geometries, gc2.geometries)
    for a, b in geoms:
        assert a.wkb == b.wkb
        assert a is not b

    pt = Point((-1, -5, -1))
    gc2[1] = pt
    assert gc1.coordinates != gc2.coordinates

    gc1.geometries[0].x = 0.5
    assert gc1.geometries[0].x == 0.5
    assert gc2.geometries[0].x == 0
    assert gc1[0].wkb != gc2[0].wkb

def test_geometrycollection_edit_wkb():
    """
    Editing object changes WKB
    """
    pt = Point((0, 0))
    gc = GeometryCollection([pt])
    wkb1 = gc.wkb

    pt.x = 1
    wkb2 = gc.wkb

    assert wkb1 != wkb2

def test_geometrycollection_index():
    """
    Overloaded [] access correct item in multigeometry
    """
    pt = Point((0, 0))
    gc = GeometryCollection([pt])

    assert gc.geometries[0] == gc[0]

def test_multigeometry_add():
    """
    Add operator for multigeometries
    """
    p1 = Point((1, 1, 1))
    p2 = Point((2, 2, 2))
    p3 = Point((3, 3, 3))
    p4 = Point((4, 4, 4))

    mp1 = MultiPoint([p1, p2])
    assert len(mp1) == 2

    mpX = mp1 + p1
    assert len(mp1) == 2
    assert len(mpX) == 3
    assert type(mpX) == MultiPoint
    
    ls = LineString([(3, 4, 5), (9, 10, 11)])

    mg = mp1 + ls
    assert len(mg) == 3
    assert type(mg) == GeometryCollection

    mp2 = MultiPoint([p3, p4])
    mp3 = mp1 + mp2
    assert mp3[0].x == 1
    assert mp3[1].x == 2
    assert mp3[2].x == 3
    assert mp3[3].x == 4
    assert len(mp3) == 4
    assert type(mp3) == MultiPoint

def test_multigeometry_iadd():
    p1 = Point((1, 1, 1))
    p2 = Point((2, 2, 2))
    p3 = Point((3, 3, 3))
    p4 = Point((4, 4, 4))
    p5 = Point((5, 5, 5))

    mp1 = MultiPoint([p1, p2])
    mp1 += p3
    assert len(mp1) == 3
    assert type(mp1) == MultiPoint

    mp2 = MultiPoint([p4, p5])
    mp1 += mp2
    assert len(mp1) == 5
    assert type(mp1) == MultiPoint

    ls = LineString([(3, 4, 5), (9, 10, 11)])
    with pytest.raises(CollectionError):
        mp1 += ls

    gc = GeometryCollection([p1, ls])
    gc += p2
    assert len(gc) == 3
    assert type(gc) == GeometryCollection

    gc += mp2
    assert len(gc) == 5
    assert type(gc) == GeometryCollection

def test_geometry_add():
    p1 = Point((1, 1, 1))
    p2 = Point((2, 2, 2))
    p3 = Point((3, 3, 3))
    p4 = Point((4, 4, 4))
    p5 = Point((5, 5, 5))

    mp1 = p1 + p2
    assert type(mp1) == MultiPoint
    assert len(mp1) == 2
    assert mp1[0].x == 1
    assert mp1[0].y == 1
    assert mp1[0].z == 1
    assert mp1[1].x == 2
    assert mp1[1].y == 2
    assert mp1[1].z == 2

    mp2 = MultiPoint([p2, p3])

    mp3 = p1 + mp2
    assert type(mp3) == MultiPoint
    assert len(mp3) == 3

    mp4 = mp2 + p1
    assert type(mp4) == MultiPoint
    assert len(mp4) == 3

    ls = LineString([(3, 4, 5), (9, 10, 11)])

def test_geometry_add_ls():
    ls1 = LineString([(3, 4, 5), (9, 10, 11)])
    ls2 = LineString([(9, 0, 0), (1, 1, 1)])

    mls = ls1 + ls2
    assert len(mls) == 2
    assert type(mls) == MultiLineString

def test_geometry_add_pg():
    geojson_pg = {"type":"Polygon","coordinates":[[[100,0],[101.0,0.0],[101.0,1.0],[100.0,1.0],[100.0,0.0]]]}
    pg1 = Geometry.from_geojson(geojson_pg, srid=4326)
    pg2 = Geometry(wkb_pg)
    pg2.srid = 4326

    mpg = pg1 + pg2
    assert len(mpg) == 2
    assert type(mpg) == MultiPolygon
    
def test_geometry_add_gc():
    p1 = Point((1, 1, 1))
    p2 = Point((2, 2, 2))
    ls = LineString([(3, 4, 5), (9, 10, 11)])
    mg1 = p1 + ls
    assert len(mg1) == 2
    assert type(mg1) == GeometryCollection

    mg2 = mg1 + p2
    assert len(mg2) == 3
    assert type(mg2) == GeometryCollection

    mg3 = p2 + mg1
    assert len(mg3) == 3
    assert type(mg3) == GeometryCollection

def test_geometry_add_srid():
    p1 = Point((1, 1, 1), srid=4326)
    p2 = Point((2, 2, 2), srid=1234)

    with pytest.raises(CollectionError):
        mp = p1 + p2

    p2.srid = 4326
    mp = p1 + p2
    assert len(mp) == 2

def test_geometry_add_srid():
    p1 = Point((0, 0), 4326)
    p2 = Point((1, 1), 4326)
    p3 = Point((2, 2), 1234)
    p4 = Point((3, 3), 1234)

    mp1 = p1 + p2
    assert mp1.srid == 4326
    mp2 = p3 + p4
    assert mp2.srid == 1234

    with pytest.raises(CollectionError):
        mp3 = mp1 + mp2

    mp1.srid = 3857
    mp2.srid = 3857
    mp3 = mp1 + mp2
    assert mp3.srid == 3857
    assert len(mp3) == 4

def test_geometry_add_srid():
    p1 = Point((0, 0))
    p2 = Point((1, 1), 4326)
    p3 = Point((2, 2))
    p4 = Point((3, 3), 1234)

    mp1 = MultiPoint([p1], 4326)
    mp1 += p2
    assert mp1.srid == 4326

    with pytest.raises(CollectionError):
        mp1 += p4

    mp2 = MultiPoint([p3], 1234)
    mp2 += p4
    assert mp2.srid == 1234

    with pytest.raises(CollectionError):
        mp3 = mp1 + mp2

def test_multigeometry_getset():
    p0 = Point((0, 0))
    p1 = Point((1, 1))
    p2 = Point((2, 2))
    mp = MultiPoint([p0, p1])
    assert len(mp) == 2
    assert mp[0].x == 0
    assert mp[0].y == 0
    assert mp[1].x == 1
    assert mp[1].y == 1

    mp[0] = p2
    assert mp[0].x == 2
    assert mp[0].y == 2
    assert mp[1].x == 1
    assert mp[1].y == 1

    ls = LineString([(3, 4), (9, 10)])
    with pytest.raises(CollectionError):
        mp[0] = ls

    gc = GeometryCollection([ls, p0])
    with pytest.raises(CollectionError):
        mp[0] = gc
    gc[0] = mp
    assert type(gc[0]) == MultiPoint

def test_interior_ring():
    p = Geometry.from_geojson(geojson_pg_ring)
    exterior = p.exterior
    interior = p.interior

    assert type(exterior) == LineString
    assert len(interior) == 1

def test_wkt_read_point():
    p = Geometry.from_wkt("POINT Z (0 1 1)")
    assert p.type == "Point"
    assert p.x == 0
    assert p.dimz == True
    assert p.dimm == False

def test_wkt_read_linestring():
    l = Geometry.from_wkt("LINESTRING (30 10, 10 30.5, 40 40) ")
    assert l.type == "LineString"
    assert l.vertices[0].x == 30
    assert l.vertices[0].y == 10
    assert l.vertices[1].x == 10
    assert l.vertices[1].y == 30.5
    assert l.dimz == False
    assert l.dimm == False

def test_wkt_read_polygon():
    p = Geometry.from_wkt("POLYGON ((99 0, 1 0, 1 1, 0 1, 0 0))")
    assert p.type == "Polygon"
    assert p.exterior.type == "LineString" 
    assert len(p.exterior.vertices) == 5
    assert p.exterior.vertices[0].x == 99
    assert p.interior == []
    assert p.dimz == False
    assert p.dimm == False

def test_wkt_read_polygon_interior():
    p = Geometry.from_wkt("POLYGON M ((7.5 1 9, 4 0 -1, 4 4 44, 0 4 0.5, 0 0 1), (1 1 -9, 1 1 2, 2 2 2, 0 2 1, 0.5 1 1))")
    assert p.type == "Polygon"
    assert p.exterior.type == "LineString" 
    assert len(p.exterior.vertices) == 5
    assert p.exterior.vertices[0].x == 7.5
    assert p.exterior.vertices[0].m == 9
    assert len(p.interior) == 1
    assert len(p.interior[0].vertices) == 5
    assert p.dimz == False
    assert p.dimm == True

def test_wkt_read_multipoint():
    mp = Geometry.from_wkt("MULTIPOINT ((0 1), (2 3))")
    assert mp.type == "MultiPoint"
    assert len(mp) == 2
    assert mp[0].x == 0
    assert mp[0].y == 1
    assert mp[1].x == 2
    assert mp[1].y == 3
    assert mp.dimz == False
    assert mp.dimm == False

def test_wkt_read_multilinestring():
    ml = Geometry.from_wkt("MULTILINESTRING ((0 0, 1 1), (2 2, 3 3))")
    assert ml.type == "MultiLineString"
    assert len(ml) == 2
    assert str(ml[0].wkb) == "01020000000200000000000000000000000000000000000000000000000000f03f000000000000f03f"
    assert ml.dimz == False
    assert ml.dimm == False

def test_wkt_read_multipolygon():
    mp = Geometry.from_wkt("MULTIPOLYGON (((1 1, 1 3, 3 3, 3 1, 1 1)), ((4 3, 6 3, 6 1, 4 1, 4 3)))  ")
    assert mp.type == "MultiPolygon"
    assert len(mp) == 2
    assert str(mp[0].wkb) == "01030000000100000005000000000000000000f03f000000000000f03f000000000000f03f0000000000000840000000000000084000000000000008400000000000000840000000000000f03f000000000000f03f000000000000f03f"
    assert mp.dimz == False
    assert mp.dimm == False

def test_wkt_read_collection():
    mp = Geometry.from_wkt("GEOMETRYCOLLECTION (MULTIPOINT((0 0), (1 1)), POINT(3 4), LINESTRING(2 3, 3 4))")
    assert mp.type == "GeometryCollection"
    assert len(mp) == 3
    assert mp[0].type == "MultiPoint"
    assert mp[1].type == "Point"
    assert mp[2].type == "LineString"
    assert mp[1].x == 3
    assert mp[1].y == 4
    assert str(mp[0].wkb) == "0104000000020000000101000000000000000000000000000000000000000101000000000000000000f03f000000000000f03f"
    assert mp.dimz == False
    assert mp.dimm == False

def test_wkt_read_collection_dimensions():
    with pytest.raises(DimensionalityError):
        mp = Geometry.from_wkt("GEOMETRYCOLLECTION (MULTIPOINT((0 0), (1 1)), POINT M (3 4 1), LINESTRING(2 3, 3 4))")

def test_ewkt_read_point():
    p = Geometry.from_wkt("SRID=4326;POINT Z (0 1 1)")
    assert p.type == "Point"
    assert p.srid == 4326
    assert p.x == 0
    assert p.dimz == True
    assert p.dimm == False

def test_ewkt_read_collection():
    mp = Geometry.from_wkt("SRID=4326;GEOMETRYCOLLECTION (MULTIPOINT((0 0), (1 1)), POINT(3 4), LINESTRING(2 3, 3 4))")
    assert mp.type == "GeometryCollection"
    assert mp.srid == 4326
    assert len(mp) == 3
    assert mp[0].type == "MultiPoint"
    assert mp[1].type == "Point"
    assert mp[2].type == "LineString"
    assert mp[1].x == 3
    assert mp[1].y == 4
    assert str(mp[0].wkb) == "0104000000020000000101000000000000000000000000000000000000000101000000000000000000f03f000000000000f03f"
    assert mp.dimz == False
    assert mp.dimm == False

def test_ewkt_read_point():
    with pytest.raises(WktError):
        p = Geometry.from_wkt("SRID=hello;POINT Z (0 1 1)")

def test_ewkt_read_empty():
    with pytest.raises(WktError):
        Geometry.from_wkt("POINT Z EMPTY")
    with pytest.raises(WktError):
        Geometry.from_wkt("LINESTRING ZM EMPTY")
    with pytest.raises(WktError):
        Geometry.from_wkt("POLYGON EMPTY")

    mg = Geometry.from_wkt("MULTIPOINT EMPTY")
    assert mg.type == "MultiPoint"
    assert len(mg) == 0

    mg = Geometry.from_wkt("MULTILINESTRING EMPTY")
    assert mg.type == "MultiLineString"
    assert len(mg) == 0

    mg = Geometry.from_wkt("MULTIPOLYGON EMPTY")
    assert mg.type == "MultiPolygon"
    assert len(mg) == 0

    mg = Geometry.from_wkt("GEOMETRYCOLLECTION EMPTY")
    assert mg.type == "GeometryCollection"
    assert len(mg) == 0
        
def test_read_wkt_malformed():
    with pytest.raises(WktError):
        p = Geometry.from_wkt("POINT(0 1 1)")

    with pytest.raises(WktError):
        p = Geometry.from_wkt("POINT ZMX (0 1 1)")

    with pytest.raises(WktError):
        p = Geometry.from_wkt("POINT EMPTY")

    with pytest.raises(WktError):
        p = Geometry.from_wkt("HELLO")

    with pytest.raises(WktError):
        p = Geometry.from_wkt("POLYGON (0 1)")

    with pytest.raises(WktError):
        p = Geometry.from_wkt("POLYGON (0 1) extra")

def test_wkt_write_empty_collection():
    mp = MultiPoint([])
    assert mp.type == "MultiPoint"
    assert len(mp) == 0
    assert mp.wkt == "MULTIPOINT EMPTY"

def test_wkt_write_point():
    wkt = "SRID=900913;POINT ZM (0 1 2 3)"
    geom = Geometry.from_wkt(wkt)
    assert geom.wkt == wkt.split(";")[1]
    assert geom.ewkt == wkt

def test_wkt_write_linestring():
    wkt = "LINESTRING (0 0, 0 1, 1 2)"
    geom = Geometry.from_wkt(wkt)
    assert geom.wkt == wkt
    
def test_wkt_write_polygon():
    wkt = "POLYGON ((0 0, 4 0, 4 4, 0 4, 0 0), (1 1, 1 2, 2 2, 2 1, 1 1))"
    geom = Geometry.from_wkt(wkt)
    assert geom.wkt == wkt
    
def test_wkt_write_multipoint():
    wkt = "MULTIPOINT ((0 0), (1 1))"
    geom = Geometry.from_wkt(wkt)
    assert geom.wkt == wkt
    
def test_wkt_write_multilinestring():
    wkt = "MULTILINESTRING ((0 0, 1 1), (2 2, 3 3))"
    geom = Geometry.from_wkt(wkt)
    assert geom.wkt == wkt
    
def test_wkt_write_multipolygon():
    wkt = "MULTIPOLYGON (((1 1, 1 3, 3 3, 3 1, 1 1)), ((4 3, 6 3, 6 1, 4 1, 4 3)), ((1 1, 1 3, 3 3, 3 1, 1 1)), ((0 0, 4 0, 4 4, 0 4, 0 0), (1 1, 1 2, 2 2, 2 1, 1 1)))"
    geom = Geometry.from_wkt(wkt)
    assert geom.wkt == wkt
    
def test_wkt_write_linestring():
    wkt = "GEOMETRYCOLLECTION (MULTIPOINT ((0 0), (1 1)), POINT (3 4), LINESTRING (2 3, 3 4))"
    geom = Geometry.from_wkt(wkt)
    assert geom.wkt == wkt
    



from .geometry import Geometry as Geometry
from .geometry import Point as Point
from .geometry import LineString as LineString
from .geometry import Polygon as Polygon
from .geometry import MultiPoint as MultiPoint
from .geometry import MultiLineString as MultiLineString
from .geometry import MultiPolygon as MultiPolygon
from .geometry import GeometryCollection as GeometryCollection
from ._version import __version__ as __version__



import numbers
from copy import copy, deepcopy
from .exceptions import (
    DependencyError,
    WkbError,
    SridError,
    DimensionalityError,
    CoordinateError,
    GeojsonError,
    CollectionError,
    WktError
)
from .hex import HexReader, HexWriter, HexBytes
from .wkt import WktReader, WktWriter

try:
    import shapely
    import shapely.wkb

    SHAPELY = True
except ImportError:
    SHAPELY = False

class Geometry:
    r"""A representation of a PostGIS geometry.

    PostGIS geometries are either an OpenGIS Consortium Simple Features for SQL
    specification type or a PostGIS extended type. The object's canonical form
    is stored in WKB or EWKB format along with an SRID and flags indicating
    whether the coordinates are 3DZ, 3DM or 4D.

    ``Geometry`` objects can be created in a number of ways. In all cases, a
    subclass for the particular geometry type will be instantiated.

    From an (E)WKB::

        >>> Geometry(b'\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')
        <Point: 'geometry(Point)'>

    From the hexadecimal string representation of an (E)WKB::

        >>> Geometry("0101000080000000000000000000000000000000000000000000000000")
        <Point: 'geometry(PointZ)'>

    The response above indicates an instance of the ``Point`` class has been
    created and that it represents a PostGIS ``geometry(PointZ)`` type.

    From a GeoJSON::

        >>> Geometry.from_geojson({'type': 'Point', 'coordinates': (0.0, 0.0)})
        <Point: 'geometry(Point,4326)'>

    From a Shapely object::

        >>> from shapely import Point
        >>> point = Point(0, 0)
        >>> Geometry.from_shapely(point, 3857)
        <Point: 'geometry(Point,3857)'>

    From any object supporting ``__geo_interface__``::

        >>> from shapefile import Reader
        >>> feature = Reader("test/multipoint.shp").shape(0)
        >>> Geometry.shape(feature)
        <MultiPoint: 'geometry(MultiPoint)'>

    A ``Geometry`` can be read as long as it is one of the following
    types: ``Point``, ``LineString``, ``Polygon``, ``MultiPoint``, ``MultiLineString``,
    ``MultiPolygon`` or ``GeometryCollection``. The M dimension will be preserved.
    """

    _WKBTYPE = 0x1FFFFFFF
    _WKBZFLAG = 0x80000000
    _WKBMFLAG = 0x40000000
    _WKBSRIDFLAG = 0x20000000
    __slots__ = ["_wkb", "_reader", "_srid", "_dimz", "_dimm"]

    def __new__(cls, wkb, srid=None, dimz=False, dimm=False):
        if cls == Geometry:
            if not wkb:
                raise WkbError("No EWKB provided")
            wkb = HexBytes(wkb)
            newcls, dimz, dimm, srid, reader = Geometry._from_wkb(wkb)
            geom = super().__new__(newcls)
            geom._wkb = wkb
            geom._reader = reader
            geom._srid = srid
            geom._dimz = dimz
            geom._dimm = dimm
        else:
            geom = super().__new__(cls)
            geom._wkb = None
            geom._reader = None
        return geom

    def __copy__(self):
        cls = self.__class__
        return cls(self.coordinates, self.srid, self.dimz, self.dimm)

    def __deepcopy__(self, _):
        return self.__copy__()

    def __add__(self, other):
        if self.srid != other.srid:
            raise CollectionError(
                    "Can not add mixed SRID types")

        if issubclass(type(other), _MultiGeometry):
            return other + self

        if type(self) is type(other):
            if type(self) is Point:
                cls = MultiPoint
            elif type(self) is LineString:
                cls = MultiLineString
            elif type(self) is Polygon:
                cls = MultiPolygon
        else:
            cls = GeometryCollection

        return cls([self, other], srid=self.srid)

    @staticmethod
    def from_geojson(geojson, srid=4326):
        """
        Create a Geometry from a GeoJSON. The SRID can be overridden from the
        expected 4326.
        """
        type_ = geojson["type"].lower()
        if type_ == "geometrycollection":
            geometries = []
            for geometry in geojson["geometries"]:
                geometries.append(Geometry.from_geojson(geometry, srid=None))
            return GeometryCollection(geometries, srid)
        if type_ == "point":
            return Point(geojson["coordinates"], srid=srid)
        if type_ == "linestring":
            return LineString(geojson["coordinates"], srid=srid)
        if type_ == "polygon":
            return Polygon(geojson["coordinates"], srid=srid)
        if type_ == "multipoint":
            geometries = _MultiGeometry._multi_from_geojson(geojson, Point)
            return MultiPoint(geometries, srid=srid)
        if type_ == "multilinestring":
            geometries = _MultiGeometry._multi_from_geojson(geojson, LineString)
            return MultiLineString(geometries, srid=srid)
        if type_ == "multipolygon":
            geometries = _MultiGeometry._multi_from_geojson(geojson, Polygon)
            return MultiPolygon(geometries, srid=srid)
        raise GeojsonError(f"Invalid GeoJSON type: {type_}")

    def _read_wkt_geom(reader):
        type_ = reader.get_type()
        dimz, dimm = reader.get_dims()

        if type_ == "POINT":
            cls = Point
        elif type_ == "LINESTRING":
            cls = LineString
        elif type_ == "POLYGON":
            cls = Polygon
        elif type_ == "MULTIPOINT":
            cls = MultiPoint
        elif type_ == "MULTILINESTRING":
            cls = MultiLineString
        elif type_ == "MULTIPOLYGON":
            cls = MultiPolygon
        elif type_ == "GEOMETRYCOLLECTION":
            cls = GeometryCollection

        return cls._read_wkt(reader)

    @staticmethod
    def from_wkt(wkt):
        reader = WktReader(wkt)
        reader.get_srid()
        geom = Geometry._read_wkt_geom(reader)
        reader.close()
        return geom

    @staticmethod
    def from_shapely(sgeom, srid=None):
        """
        Create a Geometry from a Shapely geometry and the specified SRID.

        The Shapely geometry will not be modified.
        """
        if SHAPELY:
            if srid:
                sgeom = shapely.set_srid(sgeom, srid)
            wkb_hex = shapely.to_wkb(sgeom, include_srid=True, hex=True)
            return Geometry(wkb_hex)
        raise DependencyError("Shapely")

    @staticmethod
    def shape(shape, srid=None):
        """
        Create a Geometry using ``__geo_interface__`` and the specified SRID.
        """
        return Geometry.from_geojson(shape.__geo_interface__, srid)

    @property
    def type(self):
        """
        The geometry type.
        """
        return self.__class__.__name__

    @property
    def srid(self):
        """
        The geometry SRID.
        """
        return self._srid

    @srid.setter
    def srid(self, value):
        self._check_cache()
        self._srid = value

    @property
    def coordinates(self):
        return self._coordinates()

    @property
    def geojson(self):
        """
        Get the geometry as a GeoJSON dict. There is no check that the
        GeoJSON is using an SRID of 4326.
        """
        return self._to_geojson()

    @property
    def wkb(self):
        """
        Get the geometry as an WKB.
        """
        return self._to_wkb(use_srid=False, dimz=self.dimz, dimm=self.dimm)

    @property
    def wkt(self):
        """
        Get the geometry as an WKT.
        """
        return self._to_wkt(use_srid=False)

    @property
    def ewkb(self):
        """
        Get the geometry as an EWKB.
        """
        return self._to_wkb(use_srid=True, dimz=self.dimz, dimm=self.dimm)

    @property
    def ewkt(self):
        """
        Get the geometry as an EWKT.
        """
        return self._to_wkt(use_srid=True)

    @property
    def shapely(self):
        """
        Get the geometry as a Shapely geometry. If the geometry has an SRID,
        the Shapely object will be created with it set.
        """
        return self._to_shapely()

    @property
    def bounds(self):
        """
        Get the minimum and maximum extents of the geometry: (minx, miny, maxx,
        maxy).
        """
        return self._bounds()

    @property
    def postgis_type(self):
        """
        Get the type of the geometry in PostGIS format, including additional
        dimensions and SRID if they exist.
        """
        dimz = "Z" if self.dimz else ""
        dimm = "M" if self.dimm else ""
        if self.srid:
            return f"geometry({self.type}{dimz}{dimm},{self.srid})"
        return f"geometry({self.type}{dimz}{dimm})"

    @staticmethod
    def _from_wkb(wkb):
        try:
            if wkb.startswith(b"\x00"):
                reader = HexReader(wkb, ">")  # big-endian reader
            elif wkb.startswith(b"\x01"):
                reader = HexReader(wkb, "<")  # little-endian reader
            else:
                raise WkbError("First byte in WKB must be 0 or 1.")
        except TypeError as e:
            raise WkbError() from e
        return Geometry._get_wkb_type(reader) + (reader,)

    def _check_cache(self):
        if self._reader is not None:
            self._load_geometry()
            self._wkb = None
            self._reader = None

    def _to_wkb(self, use_srid, dimz, dimm):
        if self._wkb is not None:
            return self._wkb
        writer = HexWriter("<")
        self._write_wkb_header(writer, use_srid, dimz, dimm)
        self._write_wkb(writer, dimz, dimm)
        return writer.data
    
    def _to_wkt(self, use_srid, prec=6):
        writer = WktWriter(self, use_srid, prec)
        writer.add(
            self._as_wkt(writer)
        )
        return writer.wkt

    def _to_shapely(self):
        if SHAPELY:
            sgeom = shapely.wkb.loads(self.wkb)
            srid = shapely.get_srid(sgeom)
            if srid == 0:
                srid = None
            if (srid or self.srid) and srid != self.srid:
                raise SridError(f"SRID mismatch: {srid} {self.srid}")
            return sgeom
        raise DependencyError("Shapely")

    def _to_geojson(self):
        coordinates = self._to_geojson_coordinates()
        geojson = {"type": self.type, "coordinates": coordinates}
        return geojson

    def __repr__(self):
        return f"<{self.type}: '{self.postgis_type}'>"

    def __str__(self):
        return self.ewkb.__str__()

    @property
    def __geo_interface__(self):
        return self.geojson

    def _set_dimensionality(self, geometries):
        self._dimz = None
        self._dimm = None
        for geometry in geometries:
            if self._dimz is None:
                self._dimz = geometry.dimz
            elif self._dimz != geometry.dimz:
                raise DimensionalityError("Mixed dimensionality in MultiGeometry")
            if self._dimm is None:
                self._dimm = geometry.dimm
            elif self._dimm != geometry.dimm:
                raise DimensionalityError("Mixed dimensionality in MultiGeometry")

    @staticmethod
    def _get_wkb_type(reader):
        lwgeomtype, dimz, dimm, srid = Geometry._read_wkb_header(reader)
        if lwgeomtype == 1:
            cls = Point
        elif lwgeomtype == 2:
            cls = LineString
        elif lwgeomtype == 3:
            cls = Polygon
        elif lwgeomtype == 4:
            cls = MultiPoint
        elif lwgeomtype == 5:
            cls = MultiLineString
        elif lwgeomtype == 6:
            cls = MultiPolygon
        elif lwgeomtype == 7:
            cls = GeometryCollection
        else:
            raise WkbError(f"Unsupported WKB type: {lwgeomtype}")
        return cls, dimz, dimm, srid

    @staticmethod
    def _read_wkb_header(reader):
        try:
            reader.get_char()
            header = reader.get_int()
            lwgeomtype = header & Geometry._WKBTYPE
            dimz = bool(header & Geometry._WKBZFLAG)
            dimm = bool(header & Geometry._WKBMFLAG)
            if header & Geometry._WKBSRIDFLAG:
                srid = reader.get_int()
            else:
                srid = None
        except TypeError as e:
            raise WkbError() from e
        return lwgeomtype, dimz, dimm, srid

    def _write_wkb_header(self, writer, use_srid, dimz, dimm):
        writer.add_order()
        header = (
            self._LWGEOMTYPE
            | (Geometry._WKBZFLAG if dimz else 0)
            | (Geometry._WKBMFLAG if dimm else 0)
            | (Geometry._WKBSRIDFLAG if self.srid else 0)
        )
        writer.add_int(header)
        if use_srid and self.srid:
            writer.add_int(self.srid)
        return writer

    def _as_wkt(self, writer):
        coords = self._to_wkt_coordinates(writer)
        if coords is None:
            return f"{writer.type(self)} EMPTY"
        else:
            coords = writer.wrap(coords)
            return f"{writer.type(self)} {coords}"


class _MultiGeometry(Geometry):
    __slots__ = ["_geometries"]

    def __init__(self, multitype, geometries=None, srid=None):
        if self._wkb:
            self._geometries = None
        else:
            if not all(isinstance(geometry, multitype) for geometry in geometries):
                raise CollectionError(
                    f"Found non-{multitype} when creating a Multi{multitype}"
                )
            self._geometries = geometries
            self._srid = srid
            self._set_multi_metadata()

    def __copy__(self):
        cls = self.__class__
        geometries = copy(self._geometries)
        return cls(geometries, self.srid)

    def __deepcopy__(self, _):
        cls = self.__class__
        geometries = [deepcopy(geometry) for geometry in self._geometries]
        return cls(geometries, self.srid)

    def __getitem__(self, index):
        return self._geometries[index]

    def __setitem__(self, index, value):
        if not issubclass(type(value), self.MULTITYPE):
            raise CollectionError(
                    f"Can not add {type(value)} to a {type(self)}")
        self._geometries[index] = value

    def __len__(self):
        return len(self._geometries)

    def __add__(self, other):
        if self.srid != other.srid:
            raise CollectionError(
                    "Can not add mixed SRID types")

        if type(other) is type(self):
            new_geom = copy(self)
            new_geom._geometries.extend(other._geometries)
        elif type(other) is self.MULTITYPE:
            new_geom = copy(self)
            new_geom._geometries.append(other)
        else:
            new_geom = GeometryCollection(self._geometries + [other],
                                          srid=self.srid)
        return new_geom

    def __iadd__(self, other):
        if self.srid != other.srid:
            raise CollectionError(
                    "Can not add mixed SRID types")

        if type(self) is GeometryCollection:
            if issubclass(type(other), _MultiGeometry):
                self._geometries.extend(other)
            else:
                self._geometries.append(other)
        elif type(self) is type(other):
            self._geometries.extend(other.geometries)
        elif type(other) is self.MULTITYPE:
            self._geometries.append(other)
        else:
            raise CollectionError(
                    f"Can not add a {type(other)} to a {type(self)}")
        return self

    def pop(self, index=-1):
        """
        Remove a geometry from the multigeometry.
        """
        return self._geometries.pop(index)

    @property
    def geometries(self):
        """
        List of all component geometries.
        """
        self._check_cache()
        return tuple(self._geometries)

    @property
    def dimz(self):
        """
        Whether the geometry has a Z dimension.

        :getter: ``True`` if the geometry has a Z dimension.
        :setter: Add or remove the Z dimension from this and all geometries in the collection.
        :rtype: bool
        """
        return self._dimz

    @dimz.setter
    def dimz(self, value):
        if self._dimz == value:
            return
        for geometry in self.geometries:
            geometry.dimz = value
        self._dimz = value

    @property
    def dimm(self):
        """
        Whether the geometry has an M dimension.

        :getter: ``True`` if the geometry has an M dimension.
        :setter: Add or remove the M dimension from this and all geometries in the collection.
        :rtype: bool
        """
        return self._dimm

    @dimm.setter
    def dimm(self, value):
        if self._dimm == value:
            return
        for geometry in self.geometries:
            geometry.dimm = value
        self._dimm = value

    def _bounds(self):
        bounds = [g.bounds for g in self.geometries]
        minx = min(b[0] for b in bounds)
        miny = min(b[1] for b in bounds)
        maxx = max(b[2] for b in bounds)
        maxy = max(b[3] for b in bounds)
        return (minx, miny, maxx, maxy)

    @staticmethod
    def _multi_from_geojson(geojson, cls):
        geometries = []
        for coordinates in geojson["coordinates"]:
            geometry = cls(coordinates, srid=None)
            geometries.append(geometry)
        return geometries

    def _load_geometry(self):
        self._geometries = self.__class__._read_wkb(
            self._reader, self._dimz, self._dimm
        )

    def _set_multi_metadata(self):
        self._dimz = None
        self._dimm = None
        for geometry in self.geometries:
            if self._dimz is None:
                self._dimz = geometry.dimz
            elif self._dimz != geometry.dimz:
                raise DimensionalityError("Mixed dimensionality in MultiGeometry")
            if self._dimm is None:
                self._dimm = geometry.dimm
            elif self._dimm != geometry.dimm:
                raise DimensionalityError("Mixed dimensionality in MultiGeometry")
            if geometry._srid is not None:
                if self._srid != geometry._srid:
                    raise SridError(
                        "Geometry can not be different from SRID in MultiGeometry"
                    )
                geometry._srid = None

    def _coordinates(self, dimz=True, dimm=True, tpl=True):
        return [g._coordinates(dimz, dimm, tpl) for g in self.geometries]

    def _to_geojson_coordinates(self):
        return self._coordinates(dimm=False, tpl=False)

    def _load_geometry(self):
        self._geometries = _MultiGeometry._read_wkb(
            self._reader, self._dimz, self._dimm
        )

    @staticmethod
    def _read_wkb(reader, dimz, dimm):
        geometries = []
        try:
            for _ in range(reader.get_int()):
                cls, dimz, dimm, srid = Geometry._get_wkb_type(reader)
                coordinates = cls._read_wkb(reader, dimz, dimm)
                geometry = cls(coordinates, srid=srid, dimz=dimz, dimm=dimm)
                geometries.append(geometry)
        except TypeError as e:
            raise WkbError() from e
        return geometries

    @classmethod
    def _read_wkt(cls, reader):
        if reader.get_empty():
            geoms = []
        else:
            reader.get_openpar()
            geoms = [cls.MULTITYPE._read_wkt(reader)]
            while reader.get_comma(req=False):
                geom = cls.MULTITYPE._read_wkt(reader)
                geoms.append(geom)
            reader.get_closepar()
        return cls(geoms, srid=reader.srid)

    def _write_wkb(self, writer, dimz, dimm):
        writer.add_int(len(self.geometries))
        for geometry in self._geometries:
            geometry._write_wkb_header(writer, False, dimz, dimm)
            geometry._write_wkb(writer, dimz, dimm)

    def _to_wkt_coordinates(self, writer):
        if not self.geometries:
            return None

        return writer.join(
            [writer.wrap(geom._to_wkt_coordinates(writer)) for geom in self.geometries]
        )


class Point(Geometry):
    """
    A representation of a PostGIS Point.

    ``Point`` objects can be created directly.

        >>> Point((0, -52, 5), dimm=True, srid=4326)
        <Point: 'geometry(PointM,4326)'>

    The ``dimz`` and ``dimm`` parameters will indicate how to interpret the
    coordinates that have been passed as the first argument. By default, the
    third coordinate will be interpreted as representing the Z dimension.
    """

    _LWGEOMTYPE = 1
    __slots__ = ["_x", "_y", "_z", "_m"]

    def __init__(self, coordinates=None, srid=None, dimz=False, dimm=False):
        if self._wkb:
            self._x = None
            self._y = None
            self._z = None
            self._m = None
        else:
            for c in coordinates:
                if c is None:
                    continue
                if not isinstance(c, numbers.Number):
                    raise CoordinateError(
                        f"Coordinates must be numeric: {coordinates}", coordinates
                    )
            self._srid = srid
            self._x = coordinates[0]
            self._y = coordinates[1]
            num = len(coordinates)
            if num > 4:
                raise DimensionalityError(
                    f"Maximum dimensionality supported for coordinates is 4: {coordinates}"
                )
            if num == 2:  # fill in Z and M if we are supposed to have them, else None
                if dimz and dimm:
                    self._z = 0
                    self._m = 0
                elif dimz:
                    self._z = 0
                    self._m = None
                elif dimm:
                    self._z = None
                    self._m = 0
                else:
                    self._z = None
                    self._m = None
            elif num == 3:  # use the 3rd coordinate for Z or M as directed or as Z
                if dimz and dimm:
                    self._z = coordinates[2]
                    self._m = 0
                elif dimm:
                    self._z = None
                    self._m = coordinates[2]
                else:
                    self._z = coordinates[2]
                    self._m = None
            else:  # use both the 3rd and 4th coordinates, ensure not None
                self._z = coordinates[2]
                self._m = coordinates[3]

            self._dimz = self._z is not None
            self._dimm = self._m is not None

    @property
    def x(self):
        """
        X coordinate.
        """
        self._check_cache()
        return self._x

    @x.setter
    def x(self, value):
        self._check_cache()
        self._x = value

    @property
    def y(self):
        """
        M coordinate.
        """
        self._check_cache()
        return self._y

    @y.setter
    def y(self, value):
        self._check_cache()
        self._y = value

    @property
    def z(self):
        """
        Z coordinate.
        """
        if not self._dimz:
            return None
        self._check_cache()
        return self._z

    @z.setter
    def z(self, value):
        self._check_cache()
        self._z = value
        if value is None:
            self._dimz = False
        else:
            self._dimz = True

    @property
    def m(self):
        """
        M coordinate.
        """
        if not self._dimm:
            return None
        self._check_cache()
        return self._m

    @m.setter
    def m(self, value):
        self._check_cache()
        self._m = value
        if value is None:
            self._dimm = False
        else:
            self._dimm = True

    @property
    def dimz(self):
        """
        Whether the geometry has a Z dimension.

        :getter: ``True`` if the geometry has a Z dimension.
        :setter: Add or remove the Z dimension.
        :rtype: bool
        """
        return self._dimz

    @dimz.setter
    def dimz(self, value):
        if self._dimz == value:
            return
        self._check_cache()
        if value and self._z is None:
            self._z = 0
        elif value is None:
            self._z = None
        self._dimz = value

    @property
    def dimm(self):
        """
        Whether the geometry has an M dimension.

        :getter: ``True`` if the geometry has an M dimension.
        :setter: Add or remove the M dimension.
        :rtype: bool
        """
        return self._dimm

    @dimm.setter
    def dimm(self, value):
        if self._dimm == value:
            return
        self._check_cache()
        if value and self._m is None:
            self._m = 0
        elif value is None:
            self._m = None
        self._dimm = value

    def _coordinates(self, dimz=True, dimm=True, tpl=True):
        dimz = dimz & self.dimz
        dimm = dimm & self.dimm

        if dimz and dimm:
            coordinates = (self.x, self.y, self.z, self.m)
        elif dimz:
            coordinates = (self.x, self.y, self.z)
        elif dimm:
            coordinates = (self.x, self.y, self.m)
        else:
            coordinates = (self.x, self.y)

        if tpl:
            return coordinates
        return list(coordinates)

    def _to_geojson_coordinates(self):
        return self._coordinates(dimm=False, tpl=False)

    def _bounds(self):
        return (self.x, self.y, self.x, self.y)

    def _load_geometry(self):
        self._x, self._y, self._z, self._m = Point._read_wkb(
            self._reader, self._dimz, self._dimm
        )

    @staticmethod
    def _read_wkb(reader, dimz, dimm):
        try:
            x = reader.get_double()
            y = reader.get_double()
            if dimz and dimm:
                z = reader.get_double()
                m = reader.get_double()
            elif dimz:
                z = reader.get_double()
                m = None
            elif dimm:
                z = None
                m = reader.get_double()
            else:
                z = None
                m = None
        except TypeError as e:
            raise WkbError() from e
        return x, y, z, m

    def _write_wkb(self, writer, dimz, dimm):
        writer.add_double(self.x)
        writer.add_double(self.y)
        if dimz and dimm:
            writer.add_double(self.z)
            writer.add_double(self.m)
        elif dimz:
            writer.add_double(self.z)
        elif dimm:
            writer.add_double(self.m)

    @staticmethod
    def _read_wkt_coordinates(reader):
        if reader.get_empty():
            raise WktError(reader, "Points with no coordinates are not supported in WKT.")
        reader.get_openpar()
        coords = reader.get_coordinates()
        reader.get_closepar()
        return coords

    @staticmethod
    def _read_wkt(reader):
        coords = Point._read_wkt_coordinates(reader)
        return Point(coords, dimz=reader.dimz, dimm=reader.dimm, srid=reader.srid)

    def _to_wkt_coordinates(self, writer):
        return writer.format(self.coordinates)

class LineString(Geometry):
    """
    A representation of a PostGIS Line.

    ``LineString`` objects can be created directly.

        >>> LineString([(0, 0, 0, 0), (1, 1, 0, 0), (2, 2, 0, 0)])
        <LineString: 'geometry(LineStringZM)'>

    The ``dimz`` and ``dimm`` parameters will indicate how to interpret the
    coordinates that have been passed as the first argument. By default, the
    third coordinate will be interpreted as representing the Z dimension.
    """

    _LWGEOMTYPE = 2
    __slots__ = ["_vertices"]

    def __init__(self, vertices=None, srid=None, dimz=False, dimm=False):
        if self._wkb:
            self._vertices = None
        else:
            self._srid = srid
            self._dimz = dimz
            self._dimm = dimm
            self._vertices = LineString._from_coordinates(
                vertices, dimz=dimz, dimm=dimm
            )
            self._set_dimensionality(self._vertices)

    @property
    def vertices(self):
        """
        List of vertices that comprise the line.
        """
        self._check_cache()
        return self._vertices

    @property
    def dimz(self):
        """
        Whether the geometry has a Z dimension.

        :getter: ``True`` if the geometry has a Z dimension.
        :setter: Add or remove the Z dimension from this and all vertices in the line.
        :rtype: bool
        """
        return self._dimz

    @dimz.setter
    def dimz(self, value):
        if self.dimz == value:
            return
        for vertex in self.vertices:
            vertex.dimz = value
        self._dimz = value

    @property
    def dimm(self):
        """
        Whether the geometry has a M dimension.

        :getter: ``True`` if the geometry has a M dimension.
        :setter: Add or remove the M dimension from this and all vertices in the line.
        :rtype: bool
        """
        return self._dimm

    @dimm.setter
    def dimm(self, value):
        if self.dimm == value:
            return
        for vertex in self.vertices:
            vertex.dimm = value
        self._dimm = value

    def _coordinates(self, dimz=True, dimm=True, tpl=True):
        return [v._coordinates(dimz, dimm, tpl) for v in self.vertices]

    def _to_geojson_coordinates(self):
        return self._coordinates(dimz=False, tpl=False)

    def _bounds(self):
        x = [v.x for v in self.vertices]
        y = [v.y for v in self.vertices]
        return (min(x), min(y), max(x), max(y))

    @staticmethod
    def _from_coordinates(vertices, dimz, dimm):
        return [Point(vertex, dimz=dimz, dimm=dimm) for vertex in vertices]

    def _load_geometry(self):
        vertices = LineString._read_wkb(self._reader, self._dimz, self._dimm)
        self._vertices = LineString._from_coordinates(vertices, self._dimz, self._dimm)

    @staticmethod
    def _read_wkb(reader, dimz, dimm):
        vertices = []
        try:
            for _ in range(reader.get_int()):
                coordinates = Point._read_wkb(reader, dimz, dimm)
                vertices.append(coordinates)
        except TypeError as e:
            raise WkbError() from e
        return vertices

    def _write_wkb(self, writer, dimz, dimm):
        writer.add_int(len(self.vertices))
        for vertex in self.vertices:
            vertex._write_wkb(writer, dimz, dimm)

    @staticmethod
    def _read_wkt_coordinates(reader):
        reader.get_openpar()
        vertices = [reader.get_coordinates()]
        while reader.get_comma(req=False):
            coords = reader.get_coordinates()
            vertices.append(coords)
        reader.get_closepar()
        return vertices

    @staticmethod
    def _read_wkt(reader):
        if reader.get_empty():
            raise WktError(reader, "LineStrings with no coordinates are not supported in WKT.")
        vertices = LineString._read_wkt_coordinates(reader)
        return LineString(vertices, dimz=reader.dimz, dimm=reader.dimm, srid=reader.srid)

    def _to_wkt_coordinates(self, writer):
        return writer.join(
                    [v._to_wkt_coordinates(writer) for v in self.vertices]
        )

class Polygon(Geometry):
    """
    A representation of a PostGIS Polygon.

    ``Polygon`` objects can be created directly.

        >>> Polygon([[(0, 0, 0), (1, 0, 0), (1, 1, 0), (0, 1, 0), (0, 0, 0)]])
        <Polygon: 'geometry(PolygonZ)'>

    The first polygon in the list of linear rings is the exterior ring, while
    any subsequent rings are interior boundaries.

    The ``dimz`` and ``dimm`` parameters will indicate how to interpret the
    coordinates that have been passed as the first argument. By default, the
    third coordinate will be interpreted as representing the Z dimension.
    """

    _LWGEOMTYPE = 3
    __slots__ = ["_rings"]

    def __init__(self, rings=None, srid=None, dimz=False, dimm=False):
        if self._wkb:
            self._rings = None
        else:
            self._srid = srid
            self._dimz = dimz
            self._dimm = dimm
            self._rings = Polygon._from_coordinates(rings, dimz=dimz, dimm=dimm)
            self._set_dimensionality(self._rings)

    @property
    def rings(self):
        """
        List of linearrings that comprise the polygon.
        """
        self._check_cache()
        return self._rings

    @property
    def exterior(self):
        """
        The exterior ring of the polygon.
        """
        return self.rings[0]

    @property
    def interior(self):
        """
        A list of interior rings of the polygon.
        """
        return self.rings[1:]

    @property
    def dimz(self):
        """
        Whether the geometry has a Z dimension.

        :getter: ``True`` if the geometry has a Z dimension.
        :setter: Add or remove the Z dimension from this and all linear rings in the polygon.
        :rtype: bool
        """
        return self._dimz

    @dimz.setter
    def dimz(self, value):
        if self._dimz == value:
            return
        for ring in self.rings:
            ring.dimz = value
        self._dimz = value

    @property
    def dimm(self):
        """LineString([(0, 0, 0, 0), (1, 1, 0, 0), (2, 2, 0, 0)])
        Whether the geometry has a M dimension.

        :getter: ``True`` if the geometry has a M dimension.
        :setter: Add or remove the M dimension from this and all linear rings in the polygon.
        :rtype: bool
        """
        return self._dimm

    @dimm.setter
    def dimm(self, value):
        if self._dimm == value:
            return
        for ring in self.rings:
            ring.dimm = value
        self._dimm = value

    def _coordinates(self, dimz=True, dimm=True, tpl=True):
        return [r._coordinates(dimz, dimm, tpl) for r in self.rings]

    def _to_geojson_coordinates(self):
        return self._coordinates(dimz=False, tpl=False)

    def _bounds(self):
        return self.exterior.bounds

    @staticmethod
    def _from_coordinates(rings, dimz, dimm):
        return [LineString(vertices, dimz=dimz, dimm=dimm) for vertices in rings]

    def _load_geometry(self):
        rings = Polygon._read_wkb(self._reader, self._dimz, self._dimm)
        self._rings = Polygon._from_coordinates(rings, self._dimz, self._dimm)

    @staticmethod
    def _read_wkb(reader, dimz, dimm):
        rings = []
        try:
            for _ in range(reader.get_int()):
                vertices = LineString._read_wkb(reader, dimz, dimm)
                rings.append(vertices)
        except TypeError as e:
            raise WkbError() from e
        return rings

    def _write_wkb(self, writer, dimz, dimm):
        writer.add_int(len(self.rings))
        for ring in self.rings:
            ring._write_wkb(writer, dimz, dimm)

    @staticmethod
    def _read_wkt_coordinates(reader):
        reader.get_openpar()
        rings = [LineString._read_wkt_coordinates(reader)]
        while reader.get_comma(req=False):
            ring = LineString._read_wkt_coordinates(reader)
            rings.append(ring)
        reader.get_closepar()
        return rings

    @staticmethod
    def _read_wkt(reader):
        if reader.get_empty():
            raise WktError(reader, "Polygons with no coordinates are not supported in WKT.")
        rings = Polygon._read_wkt_coordinates(reader)
        return Polygon(rings, dimz=reader.dimz, dimm=reader.dimm, srid=reader.srid)

    def _to_wkt_coordinates(self, writer):
        return writer.join(
                [writer.wrap(r._to_wkt_coordinates(writer)) for r in self.rings]
        )


class MultiPoint(_MultiGeometry):
    """
    A representation of a PostGIS MultiPoint.

    ``MultiPoint`` objects can be created directly from a list of ``Point``
    objects.

        >>> p1 = Point((0, 0, 0))
        >>> p2 = Point((1, 1, 0))
        >>> MultiPoint([p1, p2])
        <MultiPoint: 'geometry(MultiPointZ)'>

    The SRID and dimensionality of all geometries in the collection must be
    identical.
    """

    _LWGEOMTYPE = 4
    MULTITYPE = Point

    def __init__(self, points=None, srid=None):
        super().__init__(Point, geometries=points, srid=srid)

    @property
    def points(self):
        """
        List of all component points.
        """
        return self.geometries


class MultiLineString(_MultiGeometry):
    """
    A representation of a PostGIS MultiLineString

    ``MultiLineString`` objects can be created directly from a list of
    ``LineString`` objects.

        >>> l1 = LineString([(1, 1, 0), (2, 2, 0)], dimm=True)
        >>> l2 = LineString([(0, 0, 0), (0, 1, 0)], dimm=True)
        >>> MultiLineString([l1, l2])
        <MultiLineString: 'geometry(MultiLineStringM)'>

    The SRID and dimensionality of all geometries in the collection must be
    identical.
    """

    _LWGEOMTYPE = 5
    MULTITYPE = LineString

    def __init__(self, linestrings=None, srid=None):
        super().__init__(LineString, geometries=linestrings, srid=srid)

    @property
    def linestrings(self):
        """
        List of all component lines.
        """
        return self.geometries


class MultiPolygon(_MultiGeometry):
    """
    A representation of a PostGIS MultiPolygon.

    ``MultiPolygon`` objects can be created directly from a list of ``Polygon``
    objects.

        >>> p1 = Polygon([[(0, 0), (1, 0), (1, 1), (0, 1), (0, 0)]])
        >>> p2 = Polygon([[(2, 2), (3, 2), (3, 3), (2, 3), (2, 2)]])
        >>> MultiPolygon([p1, p2], srid=4326)
        <MultiPolygon: 'geometry(MultiPolygon,4326)'>

    The SRID and dimensionality of all geometries in the collection must be
    identical.
    """

    _LWGEOMTYPE = 6
    MULTITYPE = Polygon

    def __init__(self, polygons=None, srid=None):
        super().__init__(Polygon, geometries=polygons, srid=srid)

    @property
    def polygons(self):
        """
        List of all component polygons.
        """
        return self.geometries


class GeometryCollection(_MultiGeometry):
    """
    A representation of a PostGIS GeometryCollection.

    ``GeometryCollection`` objects can be created directly from a list of
    geometries, including other collections.

        >>> p = Point((0, 0, 0))
        >>> l = LineString([(1, 1, 0), (2, 2, 0)])
        >>> GeometryCollection([p, l])
        <GeometryCollection: 'geometry(GeometryCollectionZ)'>

    The SRID and dimensionality of all geometries in the collection must be
    identical.
    """

    _LWGEOMTYPE = 7
    MULTITYPE = Geometry

    def __init__(self, geometries=None, srid=None):
        super().__init__(Geometry, geometries=geometries, srid=srid)

    def _to_geojson(self):
        geometries = [g._to_geojson() for g in self._geometries]
        geojson = {"type": self.type, "geometries": geometries}
        return geojson

    def _to_wkt_coordinates(self, writer):
        return writer.join(
                    [geom._as_wkt(writer) for geom in self.geometries]
        )

    @classmethod
    def _read_wkt(cls, reader):
        if reader.get_empty():
            geoms = []
        else:
            reader.get_openpar()
            geoms = [Geometry._read_wkt_geom(reader)]
            while reader.get_comma(req=False):
                geom = Geometry._read_wkt_geom(reader)
                geoms.append(geom)
            reader.get_closepar()
        return GeometryCollection(geoms, srid=reader.srid)



import re
from .exceptions import WktError

def _regex(expr):
    return re.compile(fr"\s*{expr}\s*")

class WktReader:
    """
    A reader for Well-Knownn Text.
    """
    _TYPE = _regex("(POINT|LINESTRING|POLYGON|MULTIPOINT|MULTILINESTRING|MULTIPOLYGON|GEOMETRYCOLLECTION)")
    _DIMS = _regex("(ZM|Z|M)")
    _EMPTY = _regex("EMPTY")
    _OP = _regex("[(]")
    _CP = _regex("[)]")
    _COMMA = _regex("[,]")
    _NUMBER = _regex("[-]?[0-9]+[.]?[0-9]*")
    _SRID = _regex("SRID=[0-9]+;")

    def __init__(self, wkt, offset=0):
        self._data = wkt.upper().strip()
        self._start = offset
        self.pos = offset

    def close(self):
        """
        Terminate reading and raise error if unconsumed characters remain.
        """
        if self.pos != len(self._data):
            raise WktError(self, expected="end of WKT")

    def reset(self):
        """
        Start reading from the initial position again.
        """
        self.pos = self._start

    def _get_value(self, expr):
        match = expr.match(self._data, pos=self.pos)
        if match:
            self.pos = match.end()
            return match.group().strip()
        else:
            return None

    def _get_number(self):
        value = self._get_value(self._NUMBER)
        if not value:
            raise WktError(self, expected="number")
        return float(value)

    def get_type(self):
        value = self._get_value(self._TYPE)
        if not value:
            raise WktError(self, expected="geometry type")
        else:
            return value

    def get_dims(self):
        value = self._get_value(self._DIMS)

        if value == "Z":
            self.dimz = True
            self.dimm = False
        elif value == "M":
            self.dimz = False
            self.dimm = True
        elif value == "ZM":
            self.dimz = True
            self.dimm = True
        else:
            self.dimz = False
            self.dimm = False
        return self.dimz, self.dimm

    def get_empty(self):
        return self._get_value(self._EMPTY)

    def get_openpar(self):
        value = self._get_value(self._OP)
        if not value:
            raise WktError(self, expected="opening parenthesis")
        return True

    def get_closepar(self, req=True):
        value = self._get_value(self._CP)
        if not value:
            if req:
                raise WktError(self, expected="closing parenthesis")
            else:
                return False
        return True

    def get_srid(self):
        value = self._get_value(self._SRID)
        if value:
            value = value.strip("SRID=")
            value = value.strip(";")
            value = int(value)
        self.srid = value
        return value

    def get_comma(self, req=True):
        value = self._get_value(self._COMMA)
        if not value:
            if req:
                raise WktError(self, expected="comma")
            else:
                return False
        return True

    def get_coordinates(self):
        x = self._get_number()
        y = self._get_number()
        if self.dimz:
            z = self._get_number()
        if self.dimm:
            m = self._get_number()

        if self.dimz and self.dimm:
            return (x, y, z, m)
        elif self.dimz:
            return (x, y, z)
        elif self.dimm:
            return (x, y, m)
        else:
            return (x, y)


class WktWriter:
    """
    A writer for Well-Knownn Text.
    """
    def __init__(self, geom, use_srid=True, prec=6):
        self.geom = geom
        self.dims = False
        self.prec = prec
        self.add_srid(use_srid)
        
    def add_srid(self, use_srid):
        if use_srid and self.geom.srid:
            self.wkt = f"SRID={self.geom.srid};"
        else:
            self.wkt = ""

    def add_dims(self):
        if self.dims:
            return ""
        self.dims = True
        if self.geom.dimz and self.geom.dimm:
            return " ZM"
        elif self.geom.dimz:
            return " Z"
        elif self.geom.dimm:
            return " M"
        else:
            return ""

    def type(self, geom):
        wkt = geom.type.upper()
        wkt += self.add_dims()
        return wkt

    def add(self, text):
        self.wkt += text

    def format(self, coords):
        return " ".join([f"{c:.{self.prec}f}".rstrip("0").rstrip(".") for c in coords])

    def wrap(self, text):
        return f"({text})"

    def join(self, items):
        return ", ".join(items)
